<?php

namespace Drupal\unione\Plugin\Mail;

use Drupal\Component\Render\MarkupInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Mail\MailInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Html2Text\Html2Text;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Unione\UniOneClient;

/**
 * Default UniOne mail system plugin.
 *
 * @Mail(
 *   id = "unione_mail",
 *   label = @Translation("UniOne mailer"),
 *   description = @Translation("Sends the message using UniOne.")
 * )
 */
class UniOneMail implements MailInterface, ContainerFactoryPluginInterface {

  /**
   * A regex that matches a structure like 'Name <email@address.com>'.
   *
   * It matches anything between the first < and last > as email address.
   * This allows to use a single string to construct an Address,
   * which can be convenient to use in
   * config, and allows to have more readable config.
   * This does not try to cover all edge cases for address.
   */
  private const FROM_STRING_PATTERN = '~(?<displayName>[^<]*)<(?<addrSpec>.*)>[^>]*~';

  /**
   * The current version of the UniOne module.
   */
  private const VERSION = '1.0';

  /**
   * UniOne client.
   *
   * @var \Unione\UniOneClient
   */
  protected UniOneClient $client;

  /**
   * Configuration object.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * Logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected LoggerInterface $logger;

  /**
   * The file storage service.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $fileStorage;

  /**
   * The debug mode.
   *
   * @var bool
   */
  protected bool $isDebug;

  /**
   * UniOneMail constructor.
   *
   * @param \Drupal\Core\Config\ImmutableConfig $settings
   *   A UniOne settings config instance.
   * @param \Psr\Log\LoggerInterface $logger
   *   A logger instance.
   * @param \Drupal\Core\Entity\EntityStorageInterface $file_storage
   *   The file storage service.
   */
  public function __construct(ImmutableConfig $settings, LoggerInterface $logger, EntityStorageInterface $file_storage) {
    $this->config = $settings;
    $this->logger = $logger;
    $this->fileStorage = $file_storage;
    $this->client = new UniOneClient($this->config->get('endpoint'), $this->config->get('api_key'));
    $this->isDebug = $this->config->get('debug_mode') ?? FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $container->get('config.factory')->get('unione.settings'),
      $container->get('logger.factory')->get('unione'),
      $container->get('entity_type.manager')->getStorage('file')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function format(array $message) {
    return $message;
  }

  /**
   * {@inheritdoc}
   */
  public function mail(array $message) {
    [$body, $headers] = $this->buildMessage($message);

    $response = $this->client->send($body, $headers);
    $decode_response = Json::decode($response);

    if ($decode_response['status'] === 'success') {
      // Debug mode: log all messages.
      if ($this->isDebug) {
        $this->logger->notice('Successfully queued message from %from to %to.', [
          '%from' => $message['from'],
          '%to' => implode(', ', $decode_response['emails']),
        ]);
      }
    }

    if ($decode_response['failed_emails']) {
      $failed_emails =& $decode_response['failed_emails'];
      array_walk($failed_emails, static function (&$value, $key) {
        $value = "{$key}: {$value}";
      });

      $this->logger->error('Unable to send message from %from to: %failed_emails. %code %message', [
        '%from' => $message['from'],
        '%code' => $decode_response['code'] ? 'Error code: ' . $decode_response['code'] . '.' : '',
        '%message' => $decode_response['message'] ? 'Error message: ' . $decode_response['message'] . '.' : '',
        '%failed_emails' => $failed_emails
        ? implode(', ', $failed_emails) . '.'
        : '',
      ]);
    }

    return $decode_response['status'] === 'success';
  }

  /**
   * Builds the e-mail message in preparation to be sent to UniOne.
   *
   * @param array $message
   *   A message array, as described in hook_mail_alter().
   *   $message['params'] may contain additional parameters.
   *
   * @return array
   *   An email array formatted for UniOne delivery.
   *
   * @see https://docs.unione.io/en/web-api-ref?php#email-send
   */
  protected function buildMessage(array $message): array {
    // Add default values to make sure those array keys exist.
    $message += [
      'body' => [],
      'params' => [],
    ];

    // Build recipients list.
    foreach (explode(',', $message['to']) as $item) {
      $recipients[] = ['email' => trim($item)];
    }

    [$email, $name] = $this->checkEmail($message['headers']['From']);

    // Build the UniOne message array.
    $unione_message = [
      'from_email' => $email,
      'from_name' => $name,
      'recipients' => $recipients,
      'subject' => $message['subject'],
      'body_html' => $message['body'] instanceof MarkupInterface
      ? $message['body']->__toString()
      : $message['body'],
      'platform' => 'drupal.' . self::VERSION,
    ];

    // Remove HTML version if the message does not support HTML.
    if (isset($message['params']['html']) && !$message['params']['html']) {
      unset($unione_message['body_html']);
    }

    // Set text version of the message.
    if (isset($message['plain'])) {
      $unione_message['body_plaintext'] = $message['plain'];
    }
    else {
      $converter = new Html2Text($message['body'], ['width' => 0]);
      $unione_message['body_plaintext'] = $converter->getText();
    }

    // Add Reply-To as header according to UniOne API.
    if (!empty($message['reply-to'])) {
      [$email] = $this->checkEmail($message['reply-to']);
      $unione_message['reply_to'] = $email;
    }

    if ($this->isDebug) {
      // Left a message in watchdog with information about Cc and Bcc,
      // because they are not supported by UniOne.
      if (!empty($message['headers']['Cc'])) {
        $this->logger->notice('Cc header is not supported by UniOne.');
      }
      if (!empty($message['headers']['Bcc'])) {
        $this->logger->notice('Bcc header is not supported by UniOne.');
      }
    }

    $unione_message['track_links'] = $this->config->get('track_click');
    $unione_message['track_read'] = $this->config->get('track_read');

    // Removing elements from the header array that are static in the request.
    if (isset($message['headers']['Content-Type'])) {
      unset($message['headers']['Content-Type']);
    }
    if (isset($message['headers']['Accept'])) {
      unset($message['headers']['Accept']);
    }
    if (isset($message['headers']['X-Mailer'])) {
      $message['headers']['X-Mailer'] = 'drupal-unione';
    }

    // Make sure the files provided in the attachments array exist.
    if (!empty($message['params']['attachments'])) {
      $attachments = [];
      foreach ($message['params']['attachments'] as $attachment) {
        if (!empty($attachment['filepath']) && file_exists($attachment['filepath'])) {
          $file = $this->fileStorage->loadByProperties(['uri' => $attachment['filepath']]);
          $file = reset($file) ?: NULL;

          if ($file) {
            $attachments[] = [
              'type' => $file->getMimeType(),
              'name' => $file->getFilename(),
              'content' => base64_encode(file_get_contents($file->createFileUrl(FALSE))),
            ];
          }
        }
        elseif (!empty($attachment['filecontent']) && !empty($attachment['filename'])) {
          $attachments[] = [
            'type' => $attachment['filemime'],
            'name' => $attachment['filename'],
            'content' => base64_encode($attachment['filecontent']),
          ];
        }
      }

      if (count($attachments) > 0) {
        $unione_message['attachments'] = $attachments;
      }
    }

    return [$unione_message, $message['headers']];
  }

  /**
   * Method to split email and sender name.
   *
   * @param string $email
   *   The email address to validate.
   *
   * @return array|null
   *   The email address if it is valid, or an array of error messages.
   *
   * @see \Symfony\Component\Mime\Address::create
   */
  public function checkEmail(string $email): ?array {
    if (!str_contains($email, '<')) {
      return [$email];
    }

    if (!preg_match(self::FROM_STRING_PATTERN, $email, $matches)) {
      if ($this->isDebug) {
        $this->logger->error('Could not parse "%email" to a "%class" instance.',
          [
            '%email' => $email,
            '%class' => self::class,
          ]);
      }

      return NULL;
    }

    return [$matches['addrSpec'], $matches['displayName']];
  }

}
